from os import path
import six


def load_atoms(*args, **kwargs):
    cls = type(*args, **kwargs)
    cls.load('fireEvent')
    cls.load('getOuterHtml')
    cls.load('getInnerHtml')
    cls.load('selectText')
    return cls


@six.add_metaclass(load_atoms)
class Atoms(object):
    ATOMS = {}

    @classmethod
    def load(cls, function_name):
        filepath = path.abspath(path.join(path.dirname(__file__), 'atoms',
                                          '{}.js'.format(function_name)))
        with open(filepath, 'r') as myfile:
            cls.ATOMS[function_name] = myfile.read()

    # private

    def _execute_atom(self, function_name, *args):
        script = 'return ({}).apply(null, arguments)'.format(
            self.__class__.ATOMS.get(function_name))
        return self.driver.execute_script(script, *args)



class UserEditable(object):

    def set(self, *args):
        """
        Clear the element, the type in the given value
        :param args: value to set the input to
        """
        def func():
            self.el.clear()
            self.el.send_keys(*args)
        self._element_call(func, self.wait_for_writable)

    @property
    def value(self):
        return self.attribute_value('value')

    @value.setter  # alias
    def value(self, *args):
        self.set(*args)

    def append(self, *args):
        """
        Appends the given value to the text in the text field
        :param args: value to add to the input
        """
        self.send_keys(*args)

    def clear(self):
        """
        Clears the text field
        :return:
        """
        self._element_call(lambda: self.el.clear(), self.wait_for_writable)



import logging

import re

from ..element.selector_builder import SelectorBuilder as ElementSelectorBuilder
from ...exception import Error


class SelectorBuilder(ElementSelectorBuilder):
    def _build_wd_selector(self, selectors):
        if any(isinstance(val, re._pattern_type) for val in selectors.values()):
            return None

        if not selectors.pop('tag_name', None):
            raise Error('internal error: no tag_name?!')

        expressions = ['./tr']

        if self.query_scope.tag_name.lower() not in ['tbody', 'tfoot', 'thead']:
            expressions += ['./tbody/tr', './thead/tr', './tfoot/tr']

        attr_expr = self.xpath_builder.attribute_expression(None, selectors)

        if attr_expr:
            expressions = ['{}[{}]'.format(e, attr_expr) for e in expressions]

        xpath = " | ".join(expressions)

        logging.debug({'build_wd_selector': xpath})

        return ['xpath', xpath]



import six

from .html_elements import InputCollection
from .html_elements import Input
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class Radio(Input):
    @property
    def is_set(self):
        """
        Returns True if the element is selected
        :rtype: bool
        """
        return self._element_call(lambda: self.el.selected)

    is_selected = is_set

    def set(self):
        """ Selects the radio input """
        if not self.is_set:
            self.click()

    select = set


@six.add_metaclass(MetaHTMLElement)
class RadioCollection(InputCollection):
    # private

    @property
    def _element_class(self):
        return Radio




class Validator(object):

    @staticmethod
    def validate(element, selector):
        selector_tag_name = selector.get('tag_name')
        element_tag_name = element.tag_name.lower()

        if selector_tag_name and selector_tag_name != element_tag_name:
            return None

        return element



from copy import copy

import re
from selenium.common.exceptions import NoSuchElementException, StaleElementReferenceException
from selenium.webdriver.common.by import By

from ...exception import Error
from ...xpath_support import XpathSupport


class Locator(object):
    WD_FINDERS = {
        'class_name': By.CLASS_NAME,
        'css': By.CSS_SELECTOR,
        'id': By.ID,
        'link': By.LINK_TEXT,
        'link_text': By.LINK_TEXT,
        'name': By.NAME,
        'partial_link_text': By.PARTIAL_LINK_TEXT,
        'tag_name': By.TAG_NAME,
        'xpath': By.XPATH
    }

    # Regular expressions that can be reliably converted to xpath `contains`
    # expressions in order to optimize the .
    CONVERTABLE_REGEXP = re.compile(r'\A'
                                    r'([^\[\]\\^$.|?*+()]*)'  # leading literal characters
                                    r'[^|]*?'  # do not try to convert expressions with alternates
                                    r'([^\[\]\\^$.|?*+()]*)'  # trailing literal characters
                                    r'\Z',
                                    re.X)

    def __init__(self, query_scope, selector, selector_builder, element_validator):
        self.query_scope = query_scope  # either element or browser
        self.selector = copy(selector)
        self.selector_builder = selector_builder
        self.element_validator = element_validator

    def locate(self):
        try:
            e = self._by_id()  # short-circuit if :id is given
            if e:
                return e

            if len(self.selector) == 1:
                element = self._find_first_by_one()
            else:
                element = self._find_first_by_multiple()

            # Validation not necessary if Nerodia builds the xpath

            if 'xpath' not in list(self.selector) and 'css' not in list(self.selector):
                return element

            if element:
                return self.element_validator.validate(element, self.selector)
        except (NoSuchElementException, StaleElementReferenceException):
            return None

    def locate_all(self):
        if len(self.selector) == 1:
            return self._find_all_by_one()
        else:
            return self._find_all_by_multiple()

    # private

    def _by_id(self):
        selector = copy(self.selector)
        attr_id = selector.pop('id', None)
        if not isinstance(attr_id, str) or selector.get('adjacent'):
            return None
        tag_name = selector.pop('tag_name', None)
        if selector:  # Multiple attributes
            return None

        element = self.query_scope.wd.find_element_by_id(attr_id)
        if tag_name and not self.element_validator.validate(element, {'tag_name': tag_name}):
            return None

        return element

    def _find_first_by_one(self):
        how, what = list(self.selector.items())[0]
        self.selector_builder.check_type(how, what)

        if how in self.WD_FINDERS:
            return self._wd_find_first_by(self.WD_FINDERS.get(how), what)
        else:
            return self._find_first_by_multiple()

    def _find_first_by_multiple(self):
        selector = self.selector_builder.normalized_selector

        idx = selector.pop('index', None) if not selector.get('adjacent') else None
        visible = selector.pop('visible', None)

        built_selector = self.selector_builder.build(selector)

        if built_selector:
            # could build xpath/css for selector
            if idx is not None or visible is not None:
                idx = idx or 0
                elements = self.query_scope.wd.find_elements(*built_selector)
                if visible is not None:
                    elements = [el for el in elements if visible == el.is_displayed()]
                return elements[idx] if elements and idx < len(elements) else None
            else:
                return self.query_scope.wd.find_element(*built_selector)
        else:
            # can't use xpath, probably a regexp in there
            if idx is not None or visible is not None:
                idx = idx or 0
                elements = self._wd_find_by_regexp_selector(selector, 'select')
                if visible is not None:
                    elements = [el for el in elements if visible == el.is_displayed()]
                return elements[idx] if elements else None
            else:
                return self._wd_find_by_regexp_selector(selector, 'find')

    def _find_all_by_one(self):
        how, what = list(self.selector.items())[0]
        self.selector_builder.check_type(how, what)

        if how in self.WD_FINDERS:
            return self._wd_find_all_by(self.WD_FINDERS.get(how), what)
        else:
            return self._find_all_by_multiple()

    def _find_all_by_multiple(self):
        selector = self.selector_builder.normalized_selector
        visible = selector.pop('visible', None)

        if 'index' in selector:
            raise ValueError("can't locate all elements by index")

        built = self.selector_builder.build(selector)
        if built:
            found = self.query_scope.wd.find_elements(*built)
        else:
            found = self._wd_find_by_regexp_selector(selector, 'select')
        if visible is not None:
            return [el for el in found if visible == el.is_displayed()]
        else:
            return found

    def _wd_find_all_by(self, how, what):
        if isinstance(what, str):
            return self.query_scope.wd.find_elements(how, what)
        else:
            return [el for el in self._all_elements if what.search(self._fetch_value(el, how))]

    @staticmethod
    def _fetch_value(element, how):
        if how == 'text':
            return element.text
        elif how == 'tag_name':
            return element.tag_name.lower()
        elif how == 'href':
            href = element.get_attribute('href')
            return href and href.strip()
        else:
            return element.get_attribute(how.replace('_', '-')) or ''

    @property
    def _all_elements(self):
        return self.query_scope.wd.find_elements(By.XPATH, './/*')

    def _wd_find_first_by(self, how, what):
        if isinstance(what, str):
            return self.query_scope.wd.find_element(how, what)
        else:
            return next((x for x in self._all_elements if what.search(self._fetch_value(x, how))),
                        None)

    def _wd_find_by_regexp_selector(self, selector, method='find'):
        query_scope = self.query_scope.wd
        rx_selector = self._delete_regexps_from(selector)

        if 'label' in rx_selector and self.selector_builder.should_use_label_element:
            label = self._label_from_text(rx_selector.pop('label'))
            if not label:
                return None
            attr_id = label.get_attribute('for')
            if attr_id:
                selector['id'] = attr_id
            else:
                query_scope = label

        built_selector = self.selector_builder.build(selector)

        if not built_selector:
            raise Error('internal error: unable to build Selenium selector from'
                        ' {}'.format(selector))

        how, what = built_selector
        if how == By.XPATH and self._can_convert_regexp_to_contains:
            for key, value in rx_selector.items():
                if key == 'tag_name' or key == 'text':
                    continue

                predicates = self._regexp_selector_to_predicates(key, value)
                if predicates:
                    what = '({})[{}]'.format(what, ' and '.join(predicates))

        elements = query_scope.find_elements(how, what)
        if method == 'find':
            return next((el for el in elements if self._matches_selector(el, rx_selector)), None)
        elif method == 'select':
            return [el for el in elements if self._matches_selector(el, rx_selector)]
        else:
            return None

    @staticmethod
    def _delete_regexps_from(selector):
        rx_selector = {}

        for how, what in copy(selector).items():
            if not isinstance(what, re._pattern_type):
                continue
            rx_selector[how] = what
            selector.pop(how)

        return rx_selector

    def _label_from_text(self, label_exp):
        # TODO: this won't work correctly if @wd is a sub-element
        elements = self.query_scope.wd.find_elements(tag_name='label')
        return next((el for el in elements if self._matches_selector(el, {'text': label_exp})), None)

    def _matches_selector(self, element, selector):
        return all(what.search(self._fetch_value(element, how)) for how, what in selector.items())

    @property
    def _can_convert_regexp_to_contains(self):
        return True

    def _regexp_selector_to_predicates(self, key, regex):
        if regex.flags & re.IGNORECASE:
            return []

        match = self.CONVERTABLE_REGEXP.search(regex.pattern)
        if match is None:
            return None

        lhs = self.selector_builder.xpath_builder.lhs_for(None, key)

        return ['contains({}, {})'.format(lhs, XpathSupport.escape(group)) for
                group in match.groups() if group]



from nerodia import tag_to_class
from .elements.area import Area
from .elements.button import Button
from .elements.d_list import DList
from .elements.form import Form
from .elements.html_elements import HTMLElement, Audio, Base, Quote, Body, BR, Canvas, \
    TableCaption, TableCol, Data, DataList, Mod, Details, Dialog, Div, Embed, FieldSet, FrameSet, \
    Heading, Head, HR, Html, Input, Keygen, Label, Legend, LI, Map, Menu, MenuItem, Meta, Meter, Object, \
    OptGroup, Output, Paragraph, Param, Pre, Progress, Script, Source, Span, Style, \
    Template, TableHeaderCell, Time, Title, Track, Video, AnchorCollection, \
    HTMLElementCollection, AreaCollection, AudioCollection, BaseCollection, QuoteCollection, \
    BodyCollection, BRCollection, ButtonCollection, CanvasCollection, TableCaptionCollection, \
    TableColCollection, DataCollection, DataListCollection, ModCollection, DetailsCollection, \
    DialogCollection, DivCollection, DListCollection, EmbedCollection, FieldSetCollection, \
    FormCollection, FrameSetCollection, HeadingCollection, HeadCollection, HRCollection, \
    HtmlCollection, IFrameCollection, ImageCollection, InputCollection, KeygenCollection, \
    LabelCollection, LegendCollection, LICollection, MapCollection, MenuCollection, \
    MenuItemCollection, MetaCollection, MeterCollection, ObjectCollection, OListCollection, \
    OptGroupCollection, OptionCollection, OutputCollection, ParagraphCollection, ParamCollection, \
    PreCollection, ProgressCollection, ScriptCollection, SelectCollection, SourceCollection, \
    SpanCollection, StyleCollection, TableDataCellCollection, TemplateCollection, TextAreaCollection, \
    TableCollection, TableHeaderCellCollection, TableSectionCollection, TimeCollection, TitleCollection, \
    TableRowCollection, TrackCollection, UListCollection, VideoCollection
from .elements.i_frame import IFrame
from .elements.image import Image
from .elements.link import Anchor
from .elements.list import OList, UList
from .elements.option import Option
from .elements.select import Select
from .elements.table import Table
from .elements.table_data_cell import TableDataCell
from .elements.table_row import TableRow
from .elements.table_section import TableSection
from .elements.text_area import TextArea

tag_to_class['a'] = Anchor
tag_to_class['a_collection'] = AnchorCollection
tag_to_class['link'] = Anchor
tag_to_class['link_collection'] = AnchorCollection
tag_to_class['abbr'] = HTMLElement
tag_to_class['abbr_collection'] = HTMLElementCollection
tag_to_class['address'] = HTMLElement
tag_to_class['address_collection'] = HTMLElementCollection
tag_to_class['area'] = Area
tag_to_class['area_collection'] = AreaCollection
tag_to_class['article'] = HTMLElement
tag_to_class['article_collection'] = HTMLElementCollection
tag_to_class['aside'] = HTMLElement
tag_to_class['aside_collection'] = HTMLElementCollection
tag_to_class['audio'] = Audio
tag_to_class['audio_collection'] = AudioCollection
tag_to_class['b'] = HTMLElement
tag_to_class['b_collection'] = HTMLElementCollection
tag_to_class['base'] = Base
tag_to_class['base_collection'] = BaseCollection
tag_to_class['bdi'] = HTMLElement
tag_to_class['bdi_collection'] = HTMLElementCollection
tag_to_class['bdo'] = HTMLElement
tag_to_class['bdo_collection'] = HTMLElementCollection
tag_to_class['blockquote'] = Quote
tag_to_class['blockquote_collection'] = QuoteCollection
tag_to_class['body'] = Body
tag_to_class['body_collection'] = BodyCollection
tag_to_class['br'] = BR
tag_to_class['br_collection'] = BRCollection
tag_to_class['button'] = Button
tag_to_class['button_collection'] = ButtonCollection
tag_to_class['canvas'] = Canvas
tag_to_class['canvas_collection'] = CanvasCollection
tag_to_class['caption'] = TableCaption
tag_to_class['caption_collection'] = TableCaptionCollection
tag_to_class['cite'] = HTMLElement
tag_to_class['cite_collection'] = HTMLElementCollection
tag_to_class['code'] = HTMLElement
tag_to_class['code_collection'] = HTMLElementCollection
tag_to_class['col'] = TableCol
tag_to_class['col_collection'] = TableColCollection
tag_to_class['colgroup'] = TableCol
tag_to_class['colgroup_collection'] = TableColCollection
tag_to_class['data'] = Data
tag_to_class['data_collection'] = DataCollection
tag_to_class['datalist'] = DataList
tag_to_class['datalist_collection'] = DataListCollection
tag_to_class['dd'] = HTMLElement
tag_to_class['dd_collection'] = HTMLElementCollection
tag_to_class['del'] = Mod
tag_to_class['del_collection'] = ModCollection
tag_to_class['delete'] = Mod
tag_to_class['delete_collection'] = ModCollection
tag_to_class['details'] = Details
tag_to_class['details_collection'] = DetailsCollection
tag_to_class['dfn'] = HTMLElement
tag_to_class['dfn_collection'] = HTMLElementCollection
tag_to_class['dialog'] = Dialog
tag_to_class['dialog_collection'] = DialogCollection
tag_to_class['div'] = Div
tag_to_class['div_collection'] = DivCollection
tag_to_class['dl'] = DList
tag_to_class['dl_collection'] = DListCollection
tag_to_class['dt'] = HTMLElement
tag_to_class['dt_collection'] = HTMLElementCollection
tag_to_class['em'] = HTMLElement
tag_to_class['em_collection'] = HTMLElementCollection
tag_to_class['embed'] = Embed
tag_to_class['embed_collection'] = EmbedCollection
tag_to_class['fieldset'] = FieldSet
tag_to_class['fieldset_collection'] = FieldSetCollection
tag_to_class['figcaption'] = HTMLElement
tag_to_class['figcaption_collection'] = HTMLElementCollection
tag_to_class['figure'] = HTMLElement
tag_to_class['figure_collection'] = HTMLElementCollection
tag_to_class['footer'] = HTMLElement
tag_to_class['footer_collection'] = HTMLElementCollection
tag_to_class['form'] = Form
tag_to_class['form_collection'] = FormCollection
tag_to_class['frameset'] = FrameSet
tag_to_class['frameset_collection'] = FrameSetCollection
tag_to_class['h1'] = Heading
tag_to_class['h1_collection'] = HeadingCollection
tag_to_class['h2'] = Heading
tag_to_class['h2_collection'] = HeadingCollection
tag_to_class['h3'] = Heading
tag_to_class['h3_collection'] = HeadingCollection
tag_to_class['h4'] = Heading
tag_to_class['h4_collection'] = HeadingCollection
tag_to_class['h5'] = Heading
tag_to_class['h5_collection'] = HeadingCollection
tag_to_class['h6'] = Heading
tag_to_class['h6_collection'] = HeadingCollection
tag_to_class['head'] = Head
tag_to_class['head_collection'] = HeadCollection
tag_to_class['header'] = HTMLElement
tag_to_class['header_collection'] = HTMLElementCollection
tag_to_class['hgroup'] = HTMLElement
tag_to_class['hgroup_collection'] = HTMLElementCollection
tag_to_class['hr'] = HR
tag_to_class['hr_collection'] = HRCollection
tag_to_class['html'] = Html
tag_to_class['html_collection'] = HtmlCollection
tag_to_class['i'] = HTMLElement
tag_to_class['i_collection'] = HTMLElementCollection
tag_to_class['ital'] = HTMLElement
tag_to_class['ital_collection'] = HTMLElementCollection
tag_to_class['iframe'] = IFrame
tag_to_class['iframe_collection'] = IFrameCollection
tag_to_class['img'] = Image
tag_to_class['img_collection'] = ImageCollection
tag_to_class['input'] = Input
tag_to_class['input_collection'] = InputCollection
tag_to_class['ins'] = Mod
tag_to_class['ins_collection'] = ModCollection
tag_to_class['kbd'] = HTMLElement
tag_to_class['kbd_collection'] = HTMLElementCollection
tag_to_class['keygen'] = Keygen
tag_to_class['keygen_collection'] = KeygenCollection
tag_to_class['label'] = Label
tag_to_class['label_collection'] = LabelCollection
tag_to_class['legend'] = Legend
tag_to_class['legend_collection'] = LegendCollection
tag_to_class['li'] = LI
tag_to_class['li_collection'] = LICollection
tag_to_class['main'] = HTMLElement
tag_to_class['main_collection'] = HTMLElementCollection
tag_to_class['map'] = Map
tag_to_class['map_collection'] = MapCollection
tag_to_class['mark'] = HTMLElement
tag_to_class['mark_collection'] = HTMLElementCollection
tag_to_class['menu'] = Menu
tag_to_class['menu_collection'] = MenuCollection
tag_to_class['menuitem'] = MenuItem
tag_to_class['menuitem_collection'] = MenuItemCollection
tag_to_class['meta'] = Meta
tag_to_class['meta_collection'] = MetaCollection
tag_to_class['meter'] = Meter
tag_to_class['meter_collection'] = MeterCollection
tag_to_class['nav'] = HTMLElement
tag_to_class['nav_collection'] = HTMLElementCollection
tag_to_class['noscript'] = HTMLElement
tag_to_class['noscript_collection'] = HTMLElementCollection
tag_to_class['object'] = Object
tag_to_class['object_collection'] = ObjectCollection
tag_to_class['ol'] = OList
tag_to_class['ol_collection'] = OListCollection
tag_to_class['optgroup'] = OptGroup
tag_to_class['optgroup_collection'] = OptGroupCollection
tag_to_class['option'] = Option
tag_to_class['option_collection'] = OptionCollection
tag_to_class['output'] = Output
tag_to_class['output_collection'] = OutputCollection
tag_to_class['p'] = Paragraph
tag_to_class['p_collection'] = ParagraphCollection
tag_to_class['param'] = Param
tag_to_class['param_collection'] = ParamCollection
tag_to_class['pre'] = Pre
tag_to_class['pre_collection'] = PreCollection
tag_to_class['progress'] = Progress
tag_to_class['progress_collection'] = ProgressCollection
tag_to_class['q'] = Quote
tag_to_class['q_collection'] = QuoteCollection
tag_to_class['rp'] = HTMLElement
tag_to_class['rp_collection'] = HTMLElementCollection
tag_to_class['rt'] = HTMLElement
tag_to_class['rt_collection'] = HTMLElementCollection
tag_to_class['ruby'] = HTMLElement
tag_to_class['ruby_collection'] = HTMLElementCollection
tag_to_class['s'] = HTMLElement
tag_to_class['s_collection'] = HTMLElementCollection
tag_to_class['samp'] = HTMLElement
tag_to_class['samp_collection'] = HTMLElementCollection
tag_to_class['script'] = Script
tag_to_class['script_collection'] = ScriptCollection
tag_to_class['section'] = HTMLElement
tag_to_class['section_collection'] = HTMLElementCollection
tag_to_class['select'] = Select
tag_to_class['select_collection'] = SelectCollection
tag_to_class['select_list'] = Select
tag_to_class['select_list_collection'] = SelectCollection
tag_to_class['small'] = HTMLElement
tag_to_class['small_collection'] = HTMLElementCollection
tag_to_class['source'] = Source
tag_to_class['source_collection'] = SourceCollection
tag_to_class['span'] = Span
tag_to_class['span_collection'] = SpanCollection
tag_to_class['strong'] = HTMLElement
tag_to_class['strong_collection'] = HTMLElementCollection
tag_to_class['style'] = Style
tag_to_class['style_collection'] = StyleCollection
tag_to_class['sub'] = HTMLElement
tag_to_class['sub_collection'] = HTMLElementCollection
tag_to_class['summary'] = HTMLElement
tag_to_class['summary_collection'] = HTMLElementCollection
tag_to_class['sup'] = HTMLElement
tag_to_class['sup_collection'] = HTMLElementCollection
tag_to_class['table'] = Table
tag_to_class['table_collection'] = TableCollection
tag_to_class['tbody'] = TableSection
tag_to_class['tbody_collection'] = TableSectionCollection
tag_to_class['td'] = TableDataCell
tag_to_class['td_collection'] = TableDataCellCollection
tag_to_class['template'] = Template
tag_to_class['template_collection'] = TemplateCollection
tag_to_class['textarea'] = TextArea
tag_to_class['textarea_collection'] = TextAreaCollection
tag_to_class['tfoot'] = TableSection
tag_to_class['tfoot_collection'] = TableSectionCollection
tag_to_class['th'] = TableHeaderCell
tag_to_class['th_collection'] = TableHeaderCellCollection
tag_to_class['thead'] = TableSection
tag_to_class['thead_collection'] = TableSectionCollection
tag_to_class['time'] = Time
tag_to_class['time_collection'] = TimeCollection
tag_to_class['title'] = Title
tag_to_class['title_collection'] = TitleCollection
tag_to_class['tr'] = TableRow
tag_to_class['tr_collection'] = TableRowCollection
tag_to_class['track'] = Track
tag_to_class['track_collection'] = TrackCollection
tag_to_class['u'] = HTMLElement
tag_to_class['u_collection'] = HTMLElementCollection
tag_to_class['ul'] = UList
tag_to_class['ul_collection'] = UListCollection
tag_to_class['var'] = HTMLElement
tag_to_class['var_collection'] = HTMLElementCollection
tag_to_class['video'] = Video
tag_to_class['video_collection'] = VideoCollection
tag_to_class['wbr'] = HTMLElement
tag_to_class['wbr_collection'] = HTMLElementCollection



from ..element.locator import Locator as ElementLocator


class Locator(ElementLocator):
    def locate_all(self):
        return self._find_all_by_multiple()

    # private

    def _by_id(self):
        return None



from errno import ENOENT

import six
from os import path

from .html_elements import InputCollection
from .html_elements import Input
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class FileField(Input):
    def set(self, filepath):
        """
        Set the file field to the given path

        :param filepath: path to the file
        :raises: ENOENT
        """
        if not path.exists(filepath):
            raise ENOENT
        self._element_call(lambda: self.el.send_keys(filepath))


@six.add_metaclass(MetaHTMLElement)
class FileFieldCollection(InputCollection):
    pass



import re

import six

import nerodia
from nerodia.exception import Error, NoValueFoundException, UnknownObjectException
from nerodia.wait.wait import TimeoutError
from .html_elements import HTMLElement
from ..meta_elements import MetaHTMLElement
from ..wait.wait import Wait


@six.add_metaclass(MetaHTMLElement)
class Select(HTMLElement):
    def clear(self):
        """ Clears all selected options """
        if not self.multiple:
            raise Error('you can only clear multi-selects')

        for option in self.selected_options:
            option.click()

    def options(self, *args, **kwargs):
        """ Gets all the options in the select list
        :rtype: list[nerodia.elements.option.Option]
        """
        return self._element_call(lambda: super(Select, self).options(*args, **kwargs),
                                  self.wait_for_present)

    def includes(self, term):
        """
        Returns True if the select list has at least one option where text or label matches the
        given value
        :param term: string or regex to match against the option
        :rtype: bool
        """
        return self.option(text=term).exists or self.option(label=term).exists

    def select(self, term):
        """
        Select the option whose text or label matches the given string
        If this is a multi-select and several options match the value given, all will be selected
        :param term: string or regex to match against the option
        :return: The text of the option selected. If multiple options match, returns the first match
        :rtype: str
        """
        return self._select_by('text', term)

    def select_all(self, term):
        """
        Select all the options whose text or label matches the given string
        If this is a multi-select and several options match the value given, all will be selected
        :param term: string or regex to match against the option
        :return: The text of the first option selected.
        :rtype: str
        """
        return self._select_all_by('text', term)

    def select_value(self, value):
        """
        Selects the option(s) whose value attribute matches the given string
        If this is a multi-select and several options match the value given, all will be selected
        :param value: string or regex to match against the option
        """
        return self._select_by('value', value)

    def is_selected(self, term):
        """
        Returns True if any of the selected options' text or label matches the given value
        :param term: string or regex to match against the option
        :rtype: bool
        :raises: UnknownObjectException
        """
        by_text = self.options(text=term)
        if any(option.is_selected for option in by_text):
            return True

        by_label = self.options(label=term)
        if any(option.is_selected for option in by_label):
            return True

        if len(by_text) + len(by_label) != 0:
            return False

        raise UnknownObjectException('Unable to locate option matching {}'.format(term))

    @property
    def value(self):
        """
        Returns the value of the first selected option in the select list
        Returns None if no option is selected
        :rtype: str or None
        """
        selected = self.selected_options
        return selected[0].value if selected else None

    @property
    def text(self):
        # Returns the text of the first selected option in the select list
        # Returns nil if no option is selected
        # :rtype: str or None
        selected = self.selected_options
        return selected[0].text if selected else None

    @property
    def selected_options(self):
        """
        Returns an array of currently selected options
        :rtype: list[nerodia.elements.option.Option]
        """
        script = 'var result = [];' \
                 'var options = arguments[0].options;' \
                 'for (var i = 0; i < options.length; i++) {' \
                 '  var option = options[i];' \
                 '  if (option.selected) { result.push(option) }' \
                 '}' \
                 'return result;'
        return self._element_call(lambda: self.query_scope.execute_script(script, self.el))

    # private

    def _select_by(self, how, term):
        found = self._find_options(how, term)
        if found:
            if len(found) > 1:
                nerodia.logger.deprecate('Selecting Multiple Options with #select', '#select_all')
            return self._select_matching(found)

        raise NoValueFoundException('{} not found in select list'.format(term))

    def _select_all_by(self, how, term):
        if not self.multiple:
            raise Error('you can only use #select_all on multi-selects')

        found = self._find_options(how, term)
        if found:
            return self._select_matching(found)

        raise NoValueFoundException('{} not found in select list'.format(term))

    def _find_options(self, how, term):
        types = [six.text_type, six.binary_type, int, re._pattern_type]
        found = []

        def func(sel):
            elements = []
            if type(term) in types:
                opt = {how: term}
                elements.extend(sel.options(**opt))
                if not list(elements):
                    elements.extend(sel.options(label=term))
                if elements:
                    found.extend(elements)
                    return False
                else:
                    return not found and nerodia.relaxed_locate
            else:
                raise TypeError('expected {!r}, got {}:{}'.format(types, term, term.__class__))

        try:
            Wait.until_not(func, object=self)
            if found:
                return found
            raise NoValueFoundException('{} not found in select list'.format(term))
        except TimeoutError:
            raise NoValueFoundException('{} not found in select list'.format(term))

    def _select_matching(self, elements):
        if not self.multiple:
            elements = elements[:1]
        for element in elements:
            if not element.is_selected:
                element.click()
        return elements[0].text if elements[0].exist else ''

    def _matches_regexp(self, how, element, exp):
        if how == 'text':
            return (re.search(exp, self.el.text) or re.search(exp, self.el.label)) \
                is not None
        elif how == 'value':
            return re.search(exp, element.value) is not None
        else:
            raise Error('unknown how: {}'.format(how))



import logging


class Logger(object):
    def __init__(self):
        self._logger = logging.getLogger(__name__.split('.')[0])
        self.level = logging.WARNING
        self._filename = None

    @property
    def level(self):
        return self._logger.level

    @level.setter
    def level(self, severity):
        if type(severity) == str:
            severity = severity.upper()
        self._logger.setLevel(severity)

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, path):
        self._filename = path
        for hdlr in self._logger.handlers[:]:  # remove all old handlers
            self._logger.removeHandler(hdlr)

        if path is not None:
            fileh = logging.FileHandler(path, 'a')
            self._logger.addHandler(fileh)

    def deprecate(self, old, new):
        self.warn('[DEPRECATION] {} is deprecated. Use {} instead.'.format(old, new))

    def __getattr__(self, item):
        return getattr(self._logger, item)



from .logger import Logger

__version__ = '0.3.0'

ttc = None

tag_to_class = ttc or {}

#
# Whether or not Watip should wait for an element to be found or present before taking an action.
# Defaults to true.
#

relaxed_locate = True

#
# Default wait time for wait methods.
#

default_timeout = 30

from . import locators, tag_map  # noqa

#
# Whether the locators should be used from a different namespace.
# Defaults to nerodia.locators.
#

locator_namespace = locators

#
# Custom logger
#

logger = Logger()


def element_class_for(tag_name):
    return tag_to_class.get(tag_name)



from importlib import import_module

import re

import nerodia


class ElementCollection(object):
    def __init__(self, query_scope, selector):
        self.query_scope = query_scope
        self.selector = selector
        self.as_list = []
        self.els = []

    def __iter__(self):
        """
        Yields each element in collection

        :rtype: iter

        :Example:

        divs = browser.divs(class='kls')
        for div in divs:
             print(div.text)
        """
        for e in self.to_list:
            yield e

    def __len__(self):
        """
        Returns the number of elements in the collection
        :rtype: int
        """
        return len(self.to_list)

    def __getitem__(self, idx):
        """
        Get the element at the given index

        Also note that because of lazy loading, this will return an Element instance even if
        the index is out of bounds

        :param idx: index of wanted element, 0-indexed
        :type idx: int
        :return: instance of Element subclass
        :rtype: nerodia.elements.element.Element
        """
        try:
            return self.to_list[idx]
        except IndexError:
            return self._element_class(self.query_scope, dict(index=idx, **self.selector))

    @property
    def is_empty(self):
        """
        Returns True if no elements are found

        :Example:

        browser.select_list(name='new_user_languages').options(class_name='not_here').is_empty

        :Example:

        browser.select_list(name='new_user_languages').options(id='danish').is_empty

        :return: True if no elements are found
        :rtype: bool
        """
        return len(self) == 0

    @property
    def to_list(self):
        """
        This collection as a list
        :rtype: list[nerodia.elements.element.Element]
        """
        from .elements.html_elements import HTMLElement
        from .elements.html_elements import Input
        dic = {}
        if not self.as_list:
            elements = []
            for idx, e in enumerate(self._elements):
                element = self._element_class(self.query_scope, dict(index=idx, element=e,
                                                                     **self.selector))
                if element.__class__ in [HTMLElement, Input]:
                    element = element.to_subtype()
                    class_name = element.__class__.__name__
                    dic[class_name] = dic.get(class_name, [])
                    dic[class_name].append(element)
                    idx = len(dic[class_name]) - 1
                    new_selector = dict(self.selector, element=e, index=idx,
                                        tag_name=element.tag_name)
                    elements.append(element.__class__(self.query_scope, new_selector))
                else:
                    elements.append(element)
            self.as_list = elements
        return self.as_list

    def locate(self):
        """
        Locate all elements and return self
        :rtype: ElementCollection
        """
        self.to_list
        return self

    def __eq__(self, other):
        """
        Returns true if two element collections are equal.

        :param other: other collection
        :rtype: bool

        :Example:

        browser.select_list(name='new_user_languages').options == \
            browser.select_list(id='new_user_languages').options   #=> True

        browser.select_list(name=;new_user_role').options == \
            browser.select_list(id='new_user_languages').options   #=> false
        """
        return self.to_list == other.to_list

    eql = __eq__

    # private

    @property
    def _elements(self):
        self.query_scope._ensure_context()

        element_validator = self._element_validator_class()
        selector_builder = self._selector_builder_class(self.query_scope, self.selector,
                                                        self._element_class.ATTRIBUTES)
        locator = self._locator_class(self.query_scope, self.selector, selector_builder,
                                      element_validator)

        if not self.els:
            self.els = locator.locate_all()
        return self.els

    @property
    def _locator_class(self):
        from .locators.element.locator import Locator
        return getattr(self._import_module, 'Locator', Locator)

    @property
    def _element_validator_class(self):
        from .locators.element.validator import Validator
        return getattr(self._import_module, 'Validator', Validator)

    @property
    def _selector_builder_class(self):
        from .locators.element.selector_builder import SelectorBuilder
        return getattr(self._import_module, 'SelectorBuilder', SelectorBuilder)

    @property
    def _import_module(self):
        modules = [nerodia.locator_namespace.__name__, self._element_class_name.lower()]
        try:
            return import_module('{}.{}'.format(*modules))
        except ImportError:
            return import_module('{}.element'.format(*modules[:1]))

    @property
    def _element_class_name(self):
        return self._element_class.__name__

    @property
    def _element_class(self):
        from .elements.svg_elements import SVGElementCollection
        from .elements.html_elements import HTMLElementCollection
        name = self.__class__.__name__.replace('Collection', '')
        element_module = re.sub(r'([A-Z]{1})', r'_\1', name)[1:].lower()
        if element_module == 'frame':  # special cases
            element_module = 'i_frame'
        try:
            module = import_module('nerodia.elements.{}'.format(element_module))
        except ImportError:
            if isinstance(self, HTMLElementCollection):
                module = import_module('nerodia.elements.html_elements')
            elif isinstance(self, SVGElementCollection):
                module = import_module('nerodia.elements.svg_elements')
            else:
                raise TypeError(
                    'element class for {} could not be determined'.format(self.__class__.__name__))
        return getattr(module, name)



from copy import copy

from ..element.locator import Locator as ElementLocator
from ...elements.button import Button


class Locator(ElementLocator):
    def locate_all(self):
        return self._find_all_by_multiple()

    # private

    def _wd_find_first_by(self, how, what):
        if how == 'tag_name':
            how = 'xpath'
            what = ".//button | .//input[{}]".format(
                self.selector_builder.xpath_builder.attribute_expression('input', {
                    'type': Button.VALID_TYPES}))

        return super(Locator, self)._wd_find_first_by(how, what)

    @property
    def _can_convert_regexp_to_contains(self):
        # regexp conversion won't work with the complex xpath selector
        return False

    def _matches_selector(self, element, selector):
        if 'value' in selector:
            cpy = copy(selector)
            value = cpy.pop('value', None)

            return super(Locator, self)._matches_selector(element, cpy) and \
                (value.search(self._fetch_value(element, 'value')) or
                 value.search(self._fetch_value(element, 'text')))
        else:
            return super(Locator, self)._matches_selector(element, selector)



from copy import copy

from ..element.locator import Locator as ElementLocator
from ...elements.text_field import TextField


class Locator(ElementLocator):
    def locate_all(self):
        return self._find_all_by_multiple()

    # private

    def _wd_find_first_by(self, how, what):
        if how == 'tag_name':
            how, what = self.selector_builder._build_wd_selector({'how': what})
        return super(Locator, self)._wd_find_first_by(how, what)

    def _matches_selector(self, element, selector):
        selector = copy(selector)

        tag_name = element.tag_name.lower()

        for key in ['text', 'value', 'label']:
            if key in selector:
                correct_key = 'value' if tag_name == 'input' else 'text'
                selector[correct_key] = selector.pop(key)

        return super(Locator, self)._matches_selector(element, selector)

    def _by_id(self):
        element = super(Locator, self)._by_id()

        if element and element.get_attribute('type') not in TextField.NON_TEXT_TYPES:
            return element



from ..element.validator import Validator as ElementValidator
from ...elements.button import Button


class Validator(ElementValidator):
    @staticmethod
    def validate(element, selector):
        if element.tag_name.lower() not in ['input', 'button']:
            return None
        # TODO - Verify this is desired behavior based on
        # https://bugzilla.mozilla.org/show_bug.cgi?id=1290963
        if element.tag_name.lower() == 'input' and element.get_attribute(
                'type').lower() not in Button.VALID_TYPES:
            return None

        return element
